 -- =====================================================================
-- Copyright  2010-2011 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;   
use ieee.numeric_std.all; 
use ieee.std_logic_unsigned.all;
use work.sha3_pkg.all;	 
use work.groestl_pkg.all;

-- Groestl datapath for quasi-pipelined architecture
-- possible generics values: hs = {GROESTL_DATA_SIZE_SMALL, GROESTL_DATA_SIZE_BIG}
-- rom_style = {DISTRIBUTED, COMBINATIONAL}
-- version = {SHA3_ROUND2, SHA3_ROUND3}
-- all combinations are allowed 	 

entity groestl_datapath_pq_parallel is
generic (n	:integer := GROESTL_DATA_SIZE_SMALL; hs : integer := HASH_SIZE_256;	version	:integer:= SHA3_ROUND2;	arch:integer:= GROESTL_ARCH_PARALLEL_VER1);
port( 	 
	clk					: in std_logic; 
	rst					: in std_logic; 
	ein					: in std_logic;	
	init1				: in std_logic;	
	init2				: in std_logic;	
	init3				: in std_logic;	
	last_cycle			: in std_logic;					
	finalization		: in std_logic;	
	wr_state			: in std_logic;
	wr_result			: in std_logic;
	load_ctr			: in std_logic;										   
	wr_ctr				: in std_logic;
	p_mode				: in std_logic;
	sel_out				: in std_logic;
	eout				: in std_logic;	 
	wr_c				: in std_logic;	
	wr_seg				: in std_logic;	
	en_len				: in std_logic;	
	en_ctr				: in std_logic;	
	ls_rs_flag			: in std_logic;
	final_segment		: out std_logic;		
	last_block			: out std_logic;		
	c 					: out std_logic_vector(w-1 downto 0);
	din 				: in std_logic_vector(w-1 downto 0);
    dout 				: out std_logic_vector(w-1 downto 0));
end groestl_datapath_pq_parallel;
  
architecture parallel of groestl_datapath_pq_parallel is	   
constant zero : std_logic_vector(n-1 downto 0):= (others=>'0');  
signal  rin, from_final, to_left_round, to_right_round, from_left_register, from_right_register, to_final : std_logic_vector(n-1 downto 0); 
signal  init_value,  to_left_register, to_right_register, to_left_reg, to_right_reg, intermediate_left, two_xor, three_xor, to_init : std_logic_vector(n-1 downto 0); 
signal  din_reg : std_logic_vector(w-1 downto 0);  			 
signal msg_len_tmp : std_logic_vector(w-4-1 downto 0);
signal  ctr : std_logic_vector(3 downto 0);  
signal  round : std_logic_vector(7 downto 0);  
signal  cnt_tmp : std_logic_vector(54 downto 0);  
signal  cnt, length : std_logic_vector(w-1 downto 0);  
signal last_wire : std_logic;

begin	
	
	-- flag for final segment detection
	din_reg <= din;
	
	rd2_fs_gen: if version=SHA3_ROUND2 generate
		final_segment <= din_reg(0);		   
	end generate;
	
	rd3_fs_gen: if version=SHA3_ROUND3 generate
		final_segment <= din_reg(63);		   
	end generate;
	
	-- serial input parallel output
	shfin_gen : sipo 
	generic map ( N => n, M =>w) 
	port map (clk => clk, en => ein, input => din_reg, output => rin );
						 --to_final
	intermediate_left <= from_final when init2='1' else from_left_register; 
								 --to_final
	to_left_round <= 	(rin xor from_final) when init1='1' else intermediate_left;
	to_right_round <= 	rin  when init1='1' else from_right_register;	 
		
	-- round counter 
	rd_num : countern 
	generic map (N =>4, step=>1, style =>COUNTER_STYLE_1) 
	port map (clk=>clk, rst=>rst, load=>load_ctr, en=>wr_ctr, input=> zero(3 downto 0) ,  output=>ctr); 
	round <= zero(3 downto 0) & ctr;

	-- parallel round 	
	gr0_gen: if version=SHA3_ROUND2 generate
		left_round : entity work.groestl_pq(round2_combinational) 
			generic map (n=>n)
			port map (clk=>clk, rst=>rst, p_mode=>VCC, round=>round, input=>to_left_round, output=>to_left_register);
	
		right_round : entity work.groestl_pq(round2_combinational) 
			generic map (n=>n)
			port map (clk=>clk, rst=>rst, p_mode=>GND, round=>round, input=>to_right_round, output=>to_right_register);	
	end generate;	
		
	gr1_gen: if version=SHA3_ROUND3 generate
		left_round : entity work.groestl_pq(round3_combinational) 
			generic map (n=>n)
			port map (clk=>clk, rst=>rst, p_mode=>VCC, round=>round, input=>to_left_round, output=>to_left_register);
	
		right_round : entity work.groestl_pq(round3_combinational) 
			generic map (n=>n)
			port map (clk=>clk, rst=>rst, p_mode=>GND, round=>round, input=>to_right_round, output=>to_right_register);	
	end generate;	
		
		
				   --to_final
	to_left_reg <= from_final when last_cycle='1' else to_left_register;	
	to_right_reg <= to_right_register;
	
	-- storage register for intermediate values 	
	lstate_reg : regn 
	generic map(N=>n, init=>zero(n-1 downto 0)) 
	port map (clk => clk, rst => rst, en => wr_state, input => to_left_reg, output => from_left_register );	

	rstate_reg : regn 
	generic map(N=>n, init=>zero(n-1 downto 0)) 
	port map (clk => clk, rst => rst, en => wr_state, input => to_right_reg, output => from_right_register );		
	
	
	xor_gen_ver1: if arch=GROESTL_ARCH_PARALLEL_VER1 generate
		two_xor <= from_final xor to_left_reg; --	 from_left_register;--
		three_xor <= two_xor xor to_right_reg;--  from_right_register;--
	end generate;
	
	xor_gen_ver2: if arch=GROESTL_ARCH_PARALLEL_VER2 generate
		two_xor <= from_final xor  from_left_register;--
		three_xor <= two_xor xor from_right_register;--
	end generate;
	
		
		
	to_init <= two_xor when finalization='1' else three_xor;
	
	-- initialization vectors for different versions of Groestl
	iv224: if hs=HASH_SIZE_224 generate
		to_final <= GROESTL_INIT_VALUE_224 when init3='1' else to_init;  
	end generate;
	
	iv256: if hs=HASH_SIZE_256 generate
		to_final <= GROESTL_INIT_VALUE_256 when init3='1' else to_init;  
	end generate;  
	
	iv384: if hs=HASH_SIZE_384 generate
		to_final <= GROESTL_INIT_VALUE_384 when init3='1' else to_init;  
	end generate;  
	
	iv512: if hs=HASH_SIZE_512 generate
		to_final <= GROESTL_INIT_VALUE_512 when init3='1' else to_init;  
	end generate;  
			
	-- final message digest storage register  		
	final_reg : regn 
	generic map(N=>n, init=>zero(n-1 downto 0)) 
	port map (clk => clk, rst => rst, en => wr_result, input => to_final, output => from_final );	
	
	-- parallel input serial output	
	shfout_gen : piso 
	generic map ( N => hs , M => w ) 
	port map (clk => clk, sel => sel_out, en => eout, input => from_final(hs-1 downto 0), output => dout );
 
	rd2_ctr_gen: if CTR_SIZE=CTR_FULL generate
		-- pipelining counter - to remove critical path from this counter 
		ctr256: if n=GROESTL_DATA_SIZE_SMALL generate
			countern_gen1 : countern 
			generic map (N =>55, step=>1, style =>COUNTER_STYLE_1) 
			port map (clk=>clk, rst=>rst, load=>en_len, en=>en_ctr, input=> zero(54 downto 0) ,  output=>cnt_tmp);
		end generate;
	
		ctr512: if n=GROESTL_DATA_SIZE_BIG generate
			countern_gen1 : countern 
			generic map (N =>55, step=>2, style =>COUNTER_STYLE_1) 
			port map (clk=>clk, rst=>rst, load=>en_len, en=>en_ctr, input=> zero(54 downto 0) ,  output=>cnt_tmp);
		end generate;
		cnt <= (cnt_tmp) & "000000000";

		length <= msg_len_tmp & "0000";

		-- last block flag 	
		last_wire <= '1' when (length <= (cnt) and cnt > 0)else '0';
		last_reg : d_ff port map ( clk => clk, ena => '1', rst => '0', d => last_wire, q =>last_block);	  

		-- for controller to skip zeros in fifo	
		c <=msg_len_tmp & "0000";

		-- message len storage register 
		sob : regn 
		generic map (N=>(w-4), init=>zero(w-4-1 downto 0)) 
		port map (clk=>clk, rst=>rst, en=>wr_c, input=>din_reg(63 downto 4), output=>msg_len_tmp );
	end generate;
	
	
	rd3_ctr_gen: if CTR_SIZE=CTR_SHORT generate
		-- pipelining counter - to remove critical path from this counter 
		ctr256: if n=GROESTL_DATA_SIZE_SMALL generate
			countern_gen1 : countern 
			generic map (N =>16, step=>1, style =>COUNTER_STYLE_1) 
			port map (clk=>clk, rst=>rst, load=>en_len, en=>en_ctr, input=> zero(15 downto 0) ,  output=>cnt_tmp(15 downto 0));
		end generate;
		
		ctr512: if n=GROESTL_DATA_SIZE_BIG generate
			countern_gen1 : countern 
			generic map (N =>16, step=>2, style =>COUNTER_STYLE_1) 
			port map (clk=>clk, rst=>rst, load=>en_len, en=>en_ctr, input=> zero(15 downto 0) ,  output=>cnt_tmp(15 downto 0));
		end generate;
		cnt <= zero(38 downto 0) & cnt_tmp(15 downto 0) & zero(8 downto 0);

		length <= zero(43 downto 0) & msg_len_tmp(15 downto 0) & zero(3 downto 0);

		-- last block flag 	
		last_wire <= '1' when (length <= cnt and cnt > 0)else '0';
		last_reg : d_ff port map ( clk => clk, ena => '1', rst => '0', d => last_wire, q =>last_block);	  

		-- message len storage register 
		sob : regn 
		generic map (N=>16, init=>zero(15 downto 0)) 
		port map (clk=>clk, rst=>rst, en=>wr_c, input=>din_reg(19 downto 4), output=>msg_len_tmp(15 downto 0) );
	
		-- for controller to skip zeros in fifo	
		c <=  zero(43 downto 0) & msg_len_tmp(15 downto 0) & zero(3 downto 0);
	 end generate;

end parallel; 

